/*
 * The_xmloperator_project Software License, Version 1.7
 *
 * Copyright (c) 2000 - 2003 The_xmloperator_project.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *      "This product includes or uses software developped
 *       by The_xmloperator_project (http://www.xmloperator.net/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. Products derived from this software may not be called "xmloperator",
 *    nor may "xmloperator" appear in their name, without prior written
 *    permission. For written permission, please contact
 *    the xmloperator project administrator.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE_XMLOPERATOR_PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * Further information can be found on the project's site
 * (http://www.xmloperator.net/).
 */
package org.xmloperator.i2s;

import java.util.Iterator;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.io.PrintWriter;

import org.xmloperator.framework.document.relaxng.Name;
import org.xmloperator.framework.document.relaxng.Pattern;
import org.xmloperator.framework.document.relaxng.ElementDefinition;
import org.xmloperator.framework.document.relaxng.Schema;
import org.xmloperator.document.relaxng.impl.NameImpl;
import org.xmloperator.document.relaxng.impl.ElementDefinitionImpl;
import org.xmloperator.document.relaxng.impl.SchemaImpl;
import org.xmloperator.i2s.framework.RepeatableName;
import org.xmloperator.i2s.framework.PatternHandler;
import org.xmloperator.i2s.framework.PatternCategory;
import org.xmloperator.i2s.framework.ChoicePatternCategory;
import org.xmloperator.i2s.framework.GroupPatternCategory;
import org.xmloperator.i2s.impl.ElementPatternCategory;
import org.xmloperator.i2s.util.PatternExtractor;
import org.xmloperator.i2s.util.SchemaWriter;
import org.xmloperator.i2s.PatternCategoryFactory;

public final class SchemaGenerator implements PatternHandler {
  private static final boolean DEBUG = false;
  private static final boolean TIME = false;
  private ElementPatternCategory startPatternCategory = null;
  private final Map fromNameToPatternCategory = new HashMap();
  private Map fromNameToCount = null;
  private long time0_ms;

  public static void main(String[] args) {
    SchemaGenerator schemaGenerator = new SchemaGenerator();
    try {
      int argIndex = 0;
      while (argIndex <= args.length && args[argIndex].startsWith("-")) {
        String arg = args[argIndex++];
        if (arg.equals("-stat"))
          schemaGenerator.setStat();
        else {
          System.err.println("Unknown option \"" + arg + "\"");
          System.exit(1);
        }
      }
      for (int k = argIndex; k < args.length; k++)
        PatternExtractor.extractPatterns(
            args[k], (PatternHandler)schemaGenerator);
      schemaGenerator.convertStat();
      schemaGenerator.printSchema();
      System.exit(0);
    }
    catch (Throwable throwable) {
      throwable.printStackTrace();
      System.err.println(throwable);
      System.exit(1);
    }
  }

  private static final class Count {
    public int count = 0;
  }

  private SchemaGenerator() {
    time0_ms = System.currentTimeMillis();
  }

  private void setStat() {
    this.fromNameToCount = new HashMap();
  }

  private void convertStat() {
    if (this.fromNameToCount == null)
      return;
    Iterator iterator = this.fromNameToCount.keySet().iterator();
    while (iterator.hasNext()) {
      Name name = (Name)iterator.next();
      Count count = (Count)this.fromNameToCount.get(name);
      this.fromNameToCount.put(name, new Integer(count.count));
    }
  }

  private void printSchema() {
    if (TIME)
      System.out.println("time1_ms="
                         + (System.currentTimeMillis() - time0_ms + 1000000000));
    Map fromNameToElementDefinition = new HashMap();
    Set nameSet = this.fromNameToPatternCategory.keySet();
    ElementDefinition[] elementDefinitions =
        new ElementDefinition[nameSet.size()];
    {
      Iterator iterator = nameSet.iterator();
      int index = 0;
      while (iterator.hasNext()) {
        Name name = (Name)iterator.next();
        PatternCategory patternCategory =
            (PatternCategory)this.fromNameToPatternCategory.get(name);
        Pattern fullPattern;
        while (true) {
          if (DEBUG)
            System.out.println("Compute pattern for \""
                               + name.getName() + "\"");
          fullPattern =
              patternCategory.fullPattern(fromNameToElementDefinition);
          if (fullPattern != null)
            break;
          patternCategory =
              PatternCategoryFactory.generalizePatternCategory(patternCategory);
          fullPattern =
              patternCategory.fullPattern(fromNameToElementDefinition);
        }
        ElementDefinitionImpl elementDefinition =
            (ElementDefinitionImpl)fromNameToElementDefinition.get(name);
        if (elementDefinition == null) {
          elementDefinition = new ElementDefinitionImpl(name);
          fromNameToElementDefinition.put(name, elementDefinition);
        }
        elementDefinition.setPattern(fullPattern);
        elementDefinitions[index++] = elementDefinition;
      }
    }
    if (TIME)
      System.out.println("time2_ms="
                         + (System.currentTimeMillis() - time0_ms + 1000000000));
    SchemaImpl schema = new SchemaImpl(elementDefinitions);
    schema.setStart(
        this.startPatternCategory.fullPattern(fromNameToElementDefinition));
    PrintWriter out = new PrintWriter(System.out);
    SchemaWriter.printSchema(out, schema, this.fromNameToCount);
    if (TIME)
      System.out.println("time3_ms="
                         + (System.currentTimeMillis() - time0_ms + 1000000000));
    out.close();
  }

  ////////////////////////////////
  // PatternHandler implementation
  ////////////////////////////////

  public void pattern(
      boolean isRoot, Name containerName, Name[] attributeNames,
      boolean hasText, RepeatableName[] contentRepeatableNames) {
    if (DEBUG) {
      System.out.print("\"" + containerName.getName() + "\":");
      for (int k = 0; k < contentRepeatableNames.length; k++)
        System.out.print(" " + contentRepeatableNames[k]);
      System.out.println();
    }
    PatternCategory patternCategory =
        (PatternCategory)this.fromNameToPatternCategory.get(containerName);
    PatternCategory newPatternCategory = patternCategory != null ?
        PatternCategoryFactory.addPatternToPatternCategory(
        patternCategory, attributeNames, hasText, contentRepeatableNames) :
        PatternCategoryFactory.createPatternCategory(
        attributeNames, hasText, contentRepeatableNames);
    if (newPatternCategory != patternCategory)
      this.fromNameToPatternCategory.put(containerName, newPatternCategory);
    if (isRoot) {
      if (this.startPatternCategory == null)
        startPatternCategory = new ElementPatternCategory(containerName);
      else
        this.startPatternCategory.addElementName(containerName);
    }
    if (this.fromNameToCount != null) {
      Count count = (Count)this.fromNameToCount.get(containerName);
      if (count == null) {
        count = new Count();
        this.fromNameToCount.put(containerName, count);
      }
      count.count++;
    }
  }
}
